/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include "main.h"

typedef enum {ICACHE, DCACHE, WB} cache_source;
typedef enum {IFILL, DFILL, DWB, BUS, BUS_WB, 
	      L2ACCESS, L2FILL, L2_WB, L3ACCESS, L3FILL} cache_action;
typedef enum { C_READ, C_WRITE } c_rw;


typedef struct cent {  /* cache entry */
    unsigned tag;
    unsigned dirty:1;
    unsigned int touched;
    unsigned long available;
    int id;
    int pending;
} Cent;

typedef struct icent {  /* cache entry */
    unsigned tag;
    unsigned dirty:1;
    unsigned int touched;
    unsigned long available;
    int id;
    int pending;
} ICent;

/* the format of a cache request that branches around the
   cache event queue.
*/
typedef struct c_request {
  cache_source source;
  c_rw read_write;
  cache_action action;
  scoreboard_t *drequestor;
  context *irequestor;
  Cent *cacheline;
  int masterid;
  unsigned tag;
  long address;
  long start_time;
  long request_time;
  struct c_request *next;
} cache_request;


#define LINESIZE    64
#define LINESIZEEXP  6 
#define PGSIZE      13 /* 8 KB (in log bits) */

#define ITLB_SIZE    48
#define DTLB_SIZE    128

#define STASHASSOC   1 

#define DASSOC      2  /*1 = dm, 2 = 2-way, 4 = 4-way, etc. */
#define IASSOC      2
#define SASSOC      2
#define TASSOC      2
#define ICACHESZ  64*1024/IASSOC/LINESIZE
#define DCACHESZ  64*1024/DASSOC/LINESIZE
#define SCACHESZ  1024*1024/SASSOC/LINESIZE
#define TCACHESZ  4096*1024/TASSOC/LINESIZE
/* total cache size is LINESIZE * DASSOC * DCACHESZ */
#define ICACHEBANKS 8
#define DCACHEBANKS 8
#define L2CACHEBANKS 8
#define L3CACHEBANKS 1

#define STASHCACHESZ 1024   /* 2 ^ STASHCACHEBT */
#define STASHCACHEBT   10 
