/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

/*  This does memory address translation, from the simulated virtual
    address, to the actual location in the simulator's memory space.
    This used to be a lot safer and easier to debug, until I streamlined
    it for speed.
    It is actually used as a macro.
 */

extern mem_addr smem_address(context *current, mem_addr addr) {
    Segment s;
    unsigned long diff;

redo:
    /*
     * we do the range comparison by an unsigned subtraction
     */
    s = current->segment[(addr >> 28)];
    diff = addr - s.base;
    if(diff <= s.size)
	return (mem_addr) ((char *)s.segment + diff);
    /* 
     * expand stack if necessary -- segment[8] is always a stack segment
     * if the pointer is within a meg of sp, increase sp.
     */
    if((current->segment[8].base - addr)  < (1 * M)) {
	expand_stack(current);
	goto redo;
    }
    /* we have to set Cause and BadVAddr */
    /*
     * Let's seg fault right here
     */
    *(char *)0 = '\0';
}
