/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include "predict.h"
#include "main.h"

static int pht[PHTSIZE];
static long    btbsource[BTBSIZE/BTBASSOC][BTBASSOC],
               btbdest[BTBSIZE/BTBASSOC][BTBASSOC],
               btbstamp[BTBSIZE/BTBASSOC][BTBASSOC];
static int btbthrd[BTBSIZE/BTBASSOC][BTBASSOC],
           btbtype[BTBSIZE/BTBASSOC][BTBASSOC];


static int phthits=0, phtmisses=0, btbhits=0, btbmisses=0;

/*  in the branchtable, 0 = strongly not taken, 3 = strongly taken */
init_pht()
{

  int i;

  for (i=0; i < PHTSIZE;i++) {
    pht[i] = 2; /* weakly taken */
  }
}

init_btb()
{

  int i,j;

  for (i=0; i < BTBSIZE/BTBASSOC;i++) 
    for (j=0;j<BTBASSOC;j++) {
      btbsource[i][j] = 0;
      btbdest[i][j] = 0;
      btbstamp[i][j] = 0;
      btbtype[i][j] = 0;
      btbthrd[i][j] = 0;
    }
}

/*  Used when you are post-mispredict or post-synchronization
 *   when you want to know which direction to predict, but there
 *   isn't really a sense of right or wrong direction.
 */

int
get_bpredict(unsigned long pc, int ghr)
{
  int index;

  index = ((pc >> 2) & PHTMASK) ^ ghr;

  if ((pht[index]) > 1)
    return 1;
  else return 0;
}

/*  This get's the branch prediction, but does not
 *    update the pht, which is not updated until the
 *    instruction retires.
 *  Uses the gshare predictor.
 */

int
bpredict(unsigned long pc, int taken, int ghr)
{
  int index;

  index = ((pc >> 2) & PHTMASK) ^ ghr;

  if (taken) {
     if (pht[index] > 1) {
      phthits++;
      return PREDICT_CORRECT;
    }
    phtmisses++;
    return PREDICT_WRONG;
  }
  else { /* not taken */
    if (pht[index] < 2) {
      phthits++;
      return PREDICT_CORRECT;
    }
    phtmisses++;
    return PREDICT_WRONG;
  }
}

/* Committing instructions update the pht with
 *   their taken/not taken status.
 */

int
update_pht(unsigned long pc, int taken, int ghr)
{
  int index;

  index = ((pc >> 2) & PHTMASK) ^ ghr;

  if (taken) {
    if ( ++(pht[index]) > 3)
      pht[index] = 3;
  }
  else { /* not taken */
    if ( --(pht[index]) < 0)
      pht[index] = 0;
  }
}

extern unsigned long cyc;

/*  Does the btb lookup.  This routine updates the btb immediately,
 *   which is optimistic.  This really should be decoupled, like the
 *   pht routines.
 */

int
btblookup(unsigned long pc, unsigned long nextpc, unsigned long *wrong,
	     int thread, int taken, int type)
{
  
  int index, i, lru;
  unsigned long leaststamp;
    
  index = (pc >> 2) & BTBMASK;
  
  for (i=0; i<BTBASSOC; i++) {
    if (btbsource[index][i] == pc  && btbthrd[index][i]  == thread) {
      btbstamp[index][i] = cyc;
      if (btbdest[index][i] == nextpc || btbtype[index][i] != JUMP) { 
	btbhits++;
	return PREDICT_CORRECT;
      }
      *wrong = btbdest[index][i];
      if (taken)
	btbdest[index][i] = nextpc;
      btbmisses++;
      return PREDICT_WRONG;
    }
  }
  if (taken) { /* only put taken branches in the btb */
    leaststamp = btbstamp[index][0];
    lru = 0;
    for (i=1; i<BTBASSOC; i++) {
      if (btbstamp[index][i] < leaststamp) {
	lru = i;
	leaststamp = btbstamp[index][i];
      }
    }
    btbsource[index][lru] = pc;
    btbdest[index][lru] = nextpc;
    btbthrd[index][lru] = thread;
    btbstamp[index][lru] = cyc;
    btbtype[index][lru] = type;
  }
  else return PREDICT_CORRECT; /* fuzzy area -- it is ok if non-taken branch
				  misses in the btb */
  btbmisses++;
  if (type == JUMP) *wrong = 0;
  return PREDICT_WRONG;
}

/*   For when you want to do the btb lookup without altering
 *     the btb.
 */

unsigned long
get_btblookup(unsigned long pc, int thread)
{
  
  int index, i;
    
  index = (pc >> 2) & BTBMASK;
  
  for (i=0; i<BTBASSOC; i++) {
    if (btbsource[index][i] == pc && btbthrd[index][i]  == thread) {
      return btbdest[index][i];
    }
  }
  return 0;
}

long rs_hits=0, rs_misses=0;

predict_stats()
{
  
   printf("branch prediction, hits = %d, misses = %d, hit rate = %.2f%%\n",
	 phthits,phtmisses, (100.0*phthits/(phthits+phtmisses)));
   printf("return_stack prediction, hits = %d, misses = %d, hit rate = %.2f%%\n",
	 rs_hits, rs_misses, (100.0*rs_hits/(rs_hits+rs_misses)));
   printf("jump prediction, hits = %d, misses = %d, hit rate = %.2f%%\n",
	 btbhits,btbmisses, (100.0*btbhits/(btbhits+btbmisses)));

}

zero_pstats()
{
  
  phthits=phtmisses=rs_hits=rs_misses=btbhits=btbmisses=0;

}

/*  Return stack routines */

void
rs_push(context *current, unsigned long addr)
{
  current->return_stack[(current->rs_size+current->rs_start) % 12] = addr;
  if (current->rs_size == 12)
    current->rs_start = (current->rs_start+1) % 12;
  else
    current->rs_size++;
}

int
rs_pop(context *current, unsigned long addr, unsigned long *wrong)
{
  if (current->rs_size == 0) {
    rs_misses++;
    *wrong = 0;
    return 0;
  }
  *wrong = current->return_stack[(--current->rs_size + current->rs_start) %12];
  if (*wrong == addr) {
    rs_hits++;
    return PREDICT_CORRECT;
  }
  rs_misses++;
  return PREDICT_WRONG;
}

/* Special case for wrong-path execution.*/

unsigned long
wp_rs_pop(context *current)
{
  if (current->rs_size == 0) {
    return 0;
  }
  return current->return_stack[(--current->rs_size + current->rs_start) % 12];
}


