/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

 */

#include "stash.h"
#include "main.h"
#include "queue.h"

int i_registers_used;
int f_registers_used;

extern int numiniq;
extern int numinfq;

scoreboard_t *rename_head;
int num_in_rename;

long iqsizetotal = 0;
long fqsizetotal = 0;

extern scoreboard_t *iq_head;
extern scoreboard_t *iq_tail;
extern scoreboard_t *fq_head;
extern scoreboard_t *fq_tail;

static void
insert_in_iq(scoreboard_t *newinst)
{
  numiniq++;
  if (iq_head == NULL) {
    iq_head = iq_tail = newinst;
  }
  else {
    iq_tail->next = newinst;
    iq_tail = newinst;
  }
}

static void
insert_in_fq(scoreboard_t *newinst)
{
  numinfq++;
  if (fq_head == NULL) {
    fq_head = fq_tail = newinst;
  }
  else {
    fq_tail->next = newinst;
    fq_tail = newinst;
  }
}

int iqconf=0, fqconf=0, iregconf=0, fregconf=0;

/* Not too much interesting happens here.  Instructions
 *   get stalled only if there is no room in the instruction
 *   queues, or if there are no renaming registers
 *   available.  Register renaming is not explicitly done,
 *   but the effects are accurately modeled.  That is, I
 *   carefully track the number of renaming registers that
 *   are being used, so I know when we run out.  It is not
 *   necessary to model the renaming itself to know when we
 *   run out of registers.
 */

void
regrename()
{
  scoreboard_t *instrn;

  for (instrn = rename_head, rename_head = NULL;
       instrn!= NULL;instrn = instrn->next)
    {
      if (instrn->status & (INVALID | SQUASHED)) {
	num_in_rename--;
	continue;
      }

      /* rename registers available? */
      if (instrn->dest < 31) {
	if (i_registers_used >= I_REGS_AVAIL) {
	  rename_head = instrn;
	  iregconf++;
	  break;
	}
	else {
	  instrn->iregs_used = 1;
	  i_registers_used++;
	}
      }
      if (instrn->dest > 31 && instrn->dest != 63) {
	if (f_registers_used >= F_REGS_AVAIL) {
	  rename_head = instrn;
	  fregconf++;
	  break;
	}
	else {
	  instrn->fregs_used = 1;
	  f_registers_used++;
	}
      }

      /* queue space available? */
      if (instrn->fu == FP) {
	if (numinfq >= FQSIZE) {
	  DEBUGPRINTF("FQ full\n");
	  fqconf++;
	  rename_head = instrn;
	  f_registers_used -= instrn->fregs_used;
	  instrn->fregs_used = 0;
	  break;
	} /* space available */
	insert_in_fq(instrn);
      }
      else { /* integer queue */
	if (numiniq >= IQSIZE) {
	  DEBUGPRINTF("IQ full\n");
	  iqconf++;
	  rename_head = instrn;
	  i_registers_used -= instrn->iregs_used;
	  f_registers_used -= instrn->fregs_used;
	  instrn->iregs_used = instrn->fregs_used = 0;
	  break;
	} /* space available */
	insert_in_iq(instrn);
      }
      num_in_rename--;
    }

  if (iq_tail) iq_tail->next = NULL;
  if (fq_tail) fq_tail->next = NULL;
  iqsizetotal += numiniq;
  fqsizetotal += numinfq;
}

