/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <setjmp.h>
#include "main.h"
extern jmp_buf toplevel;
static int sanefds = 1; /* fds 0 1 2 are valid for mipsi */
static int dupin, dupout, duperr; /* saved fds */

extern void error(EType serious, char *fmt, ...) {
    char endchar;
    int oldfds;
    va_list args;

    if((oldfds = sanefds) == 0)
	restorefds();
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    if(((endchar = fmt[strlen(fmt) - 1]) != ' ') && (endchar != '\n'))
	fprintf(stderr,"\n");
    va_end(args);
    switch(serious) {
    case Fatal: exit(1);
    case TopLevel: longjmp(toplevel, 1);
    case Benign: if(oldfds == 0) restorefds(); return;
    }
}

extern void traceout(char *fmt, ...) {
    va_list args;

    if(systrace) {
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
    }
}

extern void
initfds(void)
{
    int i;

    prog_fds[0] = 0;
    prog_fds[1] = 1;
    prog_fds[2] = 2;
    for(i=3; i<MAXFDS; ++i)
	prog_fds[i] = -1;
    if(((dupin  = dup(0)) < 0) || 
       ((dupout = dup(1)) < 0) ||
       ((duperr = dup(2)) < 0))
	error(Fatal,"dup() failed in initfds()\n");
    sanefds = 0;
}

/*
 * Swaps the first three fds with the fds in variables dupin out and err.
 * Two calls leave the state unchanged and the program can continue
 * Necessary for error(Benign) to work in print messages while the program
 * is running without disturbing its file descriptors
 */
extern void restorefds(void) {
    int a, b, c;
    a = dup(0); b = dup(1); c = dup(2);
    dup2(dupin,  0);    dup2(dupout, 1);    dup2(duperr, 2);
    close(dupin); close(dupout); close(duperr);
    dupin = a; dupout = b; duperr = c;
    sanefds = 1;
}    
extern void *galloc(int size) {
    void *t;
    if((t = (void *)malloc(size)) == NULL)
	error(Fatal, "mipsi: malloc failed\n");
    return t;
}
extern void *grealloc(void *p, int size) {
    void *t;
    DEBUGPRINTF("size %d\n", size);
    if((t = (void *)realloc(p,size)) == NULL)
	error(Fatal, "mipsi: realloc failed\n");
    return t;
}
extern char *strsave(char *s) {
    char *p = (char *) galloc(strlen(s)+1);
    strcpy(p, s);
    return p;
}
/*
 * returns 0 if t matches s upto upto c characters and if
 * there is more to t, then t must match s
 */
extern int strprefix(char *s, char *t, int c) {
    while((c > 0) && (*s != '\0') && (*t != '\0') && (*s == *t))
	s++, t++, --c;
    return (c == 0) ?  ((*t == '\0') ? 0: strcmp(s,t)) : *s - *t;
}

extern void dumpregs(context *current, int regs) {
    int n;

    printf("------------------------------------------------------------\n");
    if(regs & GENREGS) {
	for(n=0; n < 32; n += 4) 
	    printf("R[%d] = 0x%0x\tR[%d] = 0x%0x\tR[%d] = 0x%0x\tR[%d] = 0x%0x\n",
		   n, current->R[n], n+1, current->R[n+1], n+2, current->R[n+2], n+3, current->R[n+3]);
    }
    if(regs & FPREGS) {
	printf("Floating point regs (float)\n");
	for(n=0; n < 32; n += 4) 
	    printf("FR[%2d] = %2.5f\tR[%2d] = %2.5f\tR[%2d] = %2.5f\tR[%2d] = %2.5f\n", n, current->FPR[n], n+1, current->FPR[n+1], n+2, current->FPR[n+2], n+3, current->FPR[n+3]);
	printf("Floating point regs (double)\n");
	for(n=0; n < 16; n += 4) 
	    printf("FR[%2d] = %2.5f\tR[%2d] = %2.5f\tR[%2d] = %2.5f\tR[%2d] = %2.5f\n",
		   2*n,current->FPR[n], 2*(n+1),current->FPR[n+1], 2*(n+2),current->FPR[n+2], 2*(n+3),current->FPR[n+3]);
    }
    printf("------------------------------------------------------------\n");
}
